//*****************************************************************************
//
// File Name	: 'programmer.c'
// Title		: Programmer Transmitter HK-T6A v2 - using Microcontroller ATMEGA8
//            		LCD Display 16x2 LCD44780 + 74HC164 using 3 wires 
// Author		: Evzen Supler - Copyright (C) 2011
// Created		: 2011.03.06
// Revised		: 2011.03.07
// Version		: 0.1
// Target MCU	: Atmel AVR Series  ATMEGA8
//
// NOTE: This code is currently below version 0.1, and therefore is considered
// to be lacking in some functionality or documentation, or may not be fully
// tested.  Nonetheless, you can expect most functions to work.
//
// This code is distributed under the GNU Public License
//		which can be found at http://www.gnu.org/licenses/gpl.txt
//
//*****************************************************************************

/*

Menu:
Button UP - move submenu to menu
Button Down - move menu to submenu
Button Left - move previous menu/submenu  and dec value
Button Right - move next menu/submenu  and inc value
Button OK State 0-3
          - 1x  - to settings value - step 1
          - 2x  - to settings value - step 5
          - 3x  - to settings value - step 25
          - 4x  - step = 0 return to submenu
State 0 move menu/submenu
State >0 setting value

Index  = MN.subMenuNo
Index  = MN.menuNo


*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "lcd44780.h"
#include "globals.h"
#include "keyboard.h"
#include "menu.h"
#include "data.h"




#define STEP  5
#define DROFSET 4
#define ENDRATEOFSET 13
#define SUBTRIMOFSET 45
#define MIX1SDOFSET 51
#define MIX1UROFSET 52
#define MIX1DROFSET 53
#define MIX1SWOFSET 54
#define MIX2OFSET 4   // MIX1xxOFSET + MIX2OFSET
#define MIX3OFSET 8
#define SWAVALOFSET 63
#define SWBVALOFSET 64
#define SWAPITADJ 65
#define SWBPITADJ 66

// define IO
#define LEDPORT  PORTB
#define LEDERR   PB0
#define LEDOK  	 PB1
#define LEDCRCOK PB2


#define HEADER   85
#define HEADER2  252
#define DOTAZ    250 
#define BAUD_RATE   115200


uint8_t State = 0;
// state kde se nachazi menu odpovida MN.menuNo
// State = 0; reim prohlizeni
// State = 1  rezim uprav + STEP
// definice menu
// uint8_t tab3 = 0b00111111;
uint8_t c,c1;
int8_t i,j,Flag,beginIndex;
// i menu j submenu
uint16_t sum;
// hodnota D/R CH1ON,CH1off ch2on ch2off ch4on ch4off

// definice dat 
// reverze   start - 3 cnt 1  data[3]
// dual rate start - 4 cnt 6  data[4-6]
// end rate  start - 13 cnt 12  data[13-24]
// subtrim   start - 45 cnt 6   data[45-50]
// mix1 sour/dest start 51 cnt 1 data[51]
// mix1 up rate start   52 cnt 1 data[52]
// mix1 down rate start 53 cnt 1 data[53]
// mix1 switch    start 54 cnt 1 data[54]
// mix2 sour/dest start 55 cnt 1 data[55]
// mix2 up rate start   56 cnt 1 data[56]
// mix2 down rate start 57 cnt 1 data[57]
// mix2 switch    start 58 cnt 1 data[58]
// mix3 sour/dest start 59 cnt 1 data[59]
// mix3 up rate start   60 cnt 1 data[60]
// mix3 down rate start 61 cnt 1 data[61]
// mix3 switch    start 62 cnt 1 data[62]
// switch fce SWA start 63 cnt 1 data[63]
// switch fce SWB start 64 cnt 1 data[64]
// var A 		  start 65 cnt 1 data[65]
// var B 		  start 66 cnt 1 data[66]
// checksum High byte	data[67]
// checksum Low byte    data[68]



// direction not left not right
enum directions direction = null;
void print_menu(void);
int8_t counter(int8_t min, int8_t max, int8_t value, int8_t step, int8_t smer);
int8_t switchval(int8_t d);   // call SWX MIXSW
int8_t incHbyte(int8_t d);    // call MIXCH
int8_t incLbyte(int8_t d);    // call MIXCH
int8_t checkcrc17(int8_t beginIndex);
int checkcrc69(void);
uint8_t USART_readByte(void);
void USART_writeByte( uint8_t c);
void init_Uart(void);
void USART_query(void);
int initRadio(void);
void SaveToEeprom(void);
void LoadFromEeprom(void);
void func101(void);
void func102(void);
void func103(void);
void func104(void);

void func201(void);
// func 202 - 206 neni potreba j

void func301(void);


void func401(void);


void func501(void);


void func601(void);
void func602(void);
void func603(void);
void func604(void);
void func605(void);

void func701(void);
void func702(void);
void func703(void);
void func704(void);
void func705(void);

void func801(void);
void func802(void);
void func803(void);
void func804(void);
void func805(void);

void func901(void);
void func1001(void);
void func1101(void);
void func1201(void);

const FuncPtr FuncPtrTable[] PROGMEM= {
func101, func102, func103,func104, 				//4 submenus of menu 1
func201, func201, func201,func201,func201,func201,  // muze byt definovana pouze jedna funkce s volanim parametru indexu submenu
// func201, func202, func203,func204,func205,func206, 	//6 submenus of menu 2
func301, func301, func301,func301,func301,func301, 	//6 submenus of menu 3
func401, func401, func401,func401,func401,func401, 	
func401, func401, func401,func401,func401,func401, 	//12 submenus of menu 4
func501, func501, func501,func501,func501,func501, 	//6 submenus of menu 5
func601, func602, func603,func604,func605, 		//5 submenus of menu 6
func701, func702, func703,func704,func705, 		//5 submenus of menu 7
func801, func802, func803,func804,func805, 		//5 submenus of menu 8
func901, 
func1001, 
func1101, 
func1201		//1 submenus of menu 9-12

};


//Menu structure
//[0] -Number of level 0 menu items
//[1]...[n] number of second level menu items
//Eg. MSTR2[1] shows that menu item 1 has 3 submenus
const uint8_t MSTR2[] PROGMEM ={
	12,	//number of menu items
	4,	//setup cnt MN101 stav setup
	6,	//of menu item 2 	stav Reverse
	6,	//of menu item 3 	stav Dr/ kanal 1,2,4 
	12,	//of menu item 4	stav End Rate CH1-6 ON OFF 
    6,  // of menu item 5   stav Subtrim
    5,  // of menu item 6
    5,  // of menu item 7
    5,  // of menu item 8
	1,  // 9
	1,  // 10
	1,  //11
	1   //12	
	}; 



uint8_t MFIndex(uint8_t mn, uint8_t sb)
// index = MFIIndex(menu, submenu)
{
	uint8_t p=0;//points to menu in table of function pointer 
	for(uint8_t i=0; i<(mn-1); i++)
	{
		p=p+pgm_read_byte(&MSTR2[i+1]);
	}
	p=p+sb-1;
	return p;
}



ISR(TIMER0_OVF_vect)
{
cli();
	// preruseni kontrola stisku tlacitek

	if (bit_is_clear(KBD_PIN, KBD_PIN_RIGHT))
	{
	   if (State==0) 
	   { 
		if (MN.subMenuNo<pgm_read_byte(&MSTR2[MN.menuNo]))
		{ 
			MN.subMenuNo++;
		}
		else
		{
			MN.subMenuNo=1;
		}

		} // State=0
	
		direction = plus;
		print_menu();
		loop_until_bit_is_set(KBD_PIN, KBD_PIN_RIGHT);


  	}

	if (bit_is_clear(KBD_PIN, KBD_PIN_LEFT))
	{
		if (State==0)
		{
		if (MN.subMenuNo==1)
		{ 
			MN.subMenuNo=pgm_read_byte(&MSTR2[MN.menuNo]);
		}
		else
		{
			MN.subMenuNo--;
		}
		} // State=0
				
		direction = minus;
		print_menu();
		loop_until_bit_is_set(KBD_PIN, KBD_PIN_LEFT);
	}

if (bit_is_clear(KBD_PIN, KBD_PIN_DOWN))
	{
		direction = null;
		// cancel State > 0;
		if (State>0) State =0;

		if (MN.menuNo==1)
		{ 
			MN.menuNo=pgm_read_byte(&MSTR2[0]);
			MN.subMenuNo=1;
		}
		else
		{
			MN.menuNo--;
		}
		print_menu();
		loop_until_bit_is_set(KBD_PIN, KBD_PIN_DOWN);
	}
if (bit_is_clear(KBD_PIN, KBD_PIN_UP))
	{
	   	direction = null;
		if (MN.menuNo<pgm_read_byte(&MSTR2[0]))
		{ 
			MN.menuNo++;
			MN.subMenuNo=1;
		}
		else
		{
			MN.menuNo=1;
		}
		print_menu();

		loop_until_bit_is_set(KBD_PIN, KBD_PIN_DOWN);
	}

if (bit_is_clear(KBD_PIN, KBD_PIN_OK))
	{
	    // set step 
		// 1x OK = step 1
		// 2x OK = step 5
		// 3x OK = step 25
		// 4x OK = step 0 
		if (State == 0)
		{ 
		   State++;
		}
		else 
		{
		  State = State * STEP;
		}
		if (State > 25) State = 0;
		direction = null;
		print_menu();
		loop_until_bit_is_set(KBD_PIN, KBD_PIN_OK);
	}
 
sei();
}


void print_menu()
{
		lcd_clear();
		lcd_puts_pxy(0,0,MENU[MN.menuNo-1]);
		lcd_puts_pxy(0,1,SUBMENU[MFIndex(MN.menuNo, MN.subMenuNo)]);
		FPtr=(FuncPtr)pgm_read_word(&FuncPtrTable[MFIndex(MN.menuNo, MN.subMenuNo)]);	
   //     Flag = 0;
		FPtr();		

}




// void func_reverze(int8_t bit)
// parametr bit pro volani func_reverze 
// je index submenu 
// {
//   if (bit_is_clear(tab3,(bit-1))) 
//     lcd_send_char('N');
//  else lcd_send_char('R');
// }

void func101(void)
// load from eeprom
{
//  lcd_putintxy(0,13,State,2);
  if (State >0)
  {	
  LoadFromEeprom();
  State = 0;
  lcd_puts_p(MNOK);
  }
}
void func102(void)
// save to eeprom
{
  if (State >0)
  {	
  SaveToEeprom();
  State = 0;
  lcd_puts_p(MNOK);
  }

}

void func103(void)
// load data from TRX

//    lcd_putintxy(13,0,State,2);
{
    if (State > 0)
	{
	cli();
    LEDPORT &= ~(1<<LEDERR);
    LEDPORT &= ~(1<<LEDCRCOK);

//	lcd_putintxy(13,0,State,2);
	i=0;
    while(1)
	{
	 c= USART_readByte(); 
	 if (c == HEADER) 
	 {
	 data[0] = c;
     // CTI DALSI BYTE A POKUD BUDE FD BREAK
     	 c = USART_readByte();
    	 data[1] = c;
   		 if (c == 0xFD) break;
//     break;	 
	 }
      i++;
	  if (i> 10) 
	  {
	    i=0;
		USART_query();
	  } 
	} // while(1) 
    for (i=2; i<69; i++)
	{
      c = USART_readByte();
		  data[i] = c ;
	  //  eeprom_write_byte((uint8_t*)i,c);  
	}  
    
	 
	if (checkcrc69() == 0) 
	{
	  	LEDPORT |= (1<< LEDCRCOK); 
	}
    else 
	{
		LEDPORT |= (1<<LEDERR); 
	}
	sei();
    State =0;
	lcd_puts_p(MNOK);
    } // end if state >0 
	// lcd_send_char('R');
}

void func104(void)
// save to Transmitter
{
  if (State>0)
  {

  // modifi checksum
  sum = 0;
  cli();
  for (i=2; i<67; i++)
  {
	sum = sum + (uint8_t)data[i];
  }
  // hight byte
  data[67] = sum /256;
  //  data[67] = data[67] +1 ; 
  // low byte
  data[68] = sum % 256;
  // write to uart
  data[1] = 255;
  for (i=0; i<69; i++)
  {
  		c= data[i];
		  USART_writeByte(c); 
  } 

  lcd_putintxy(13,0,data[67],2);
  lcd_putintxy(12,1,data[68],4);

//  lcd_puts_p(MNOK);
  State=0; 
  sei();
  }
}
void func201(void)  // reverze CH1
{
// int8_t p =	MN.subMenuNo-1 ;
if (State > 0)
  {
	 State=0;  // shodit enter
	 data[3] ^= (1<<(MN.subMenuNo-1));
  }
	if (bit_is_clear(data[3],MN.subMenuNo-1)) lcd_send_char('N');
  else lcd_send_char('R');

}

// dual rate ch1, ch2, ch4
void func301(void)
{

    data[DROFSET+MN.subMenuNo-1] = counter(0, +100, data[DROFSET+MN.subMenuNo-1], State, direction);  
    lcd_gotoxy(8,1);
	lcd_putint((data[DROFSET+MN.subMenuNo-1]),3);
//  lcd_putintxy(13,0,State,2);
    
}



// end rate
void func401(void)
{

    data[ENDRATEOFSET+MN.subMenuNo-1] = counter(0, +120, data[ENDRATEOFSET+MN.subMenuNo-1], State, direction);  
    lcd_gotoxy(5,2);  
	lcd_putint((data[ENDRATEOFSET+MN.subMenuNo-1]),3);
//  lcd_putintxy(13,0,State,2);


}


//subtrim
void func501(void)
{
    data[SUBTRIMOFSET+MN.subMenuNo-1] = counter(-120, +120, data[SUBTRIMOFSET+MN.subMenuNo-1], State, direction);  
    lcd_gotoxy(4,1);  
    lcd_putint((data[SUBTRIMOFSET+MN.subMenuNo-1]),4);
	if (data[SUBTRIMOFSET+MN.subMenuNo-1] > 0)
	{
      lcd_gotoxy(4,1);
      lcd_send_char('+'); 

	}
	lcd_gotoxy(7,1);

}



void func601(void)
{
// MIX1 tab[51]  high nible=source
 
   	data[MIX1SDOFSET] = incHbyte(data[MIX1SDOFSET]);  
    
  	lcd_puts_p(MIXCH[data[MIX1SDOFSET]/16]);
 	// lcd_putintxy(12,0,data[51],2);

}

void func602(void)
// MIX1 tab[51] low nible=destinat 
{
    	  
   	data[MIX1SDOFSET] = incLbyte(data[MIX1SDOFSET]);	
 	lcd_puts_p(MIXCH[data[MIX1SDOFSET] & 0x0F]);

}

void func603(void)
// tab[52] MIX1 Up rate
{
  
  data[MIX1UROFSET] = counter(-100, +100, data[MIX1UROFSET], State, direction); 
  lcd_gotoxy(9,1); 
  lcd_putint(data[MIX1UROFSET],4);
  
}

void func604(void)
//  tab[53] MIX1 Down rate
{
  data[MIX1DROFSET] = counter(-100, +100, data[MIX1DROFSET], State, direction); 
  lcd_gotoxy(9,1); 
  lcd_putint(data[MIX1DROFSET],4);
}

void func605(void)
{

    data[MIX1SWOFSET] = switchval(data[MIX1SWOFSET]);
	lcd_puts_p(MIXSW[data[MIX1SWOFSET]]);
}


void func701(void)
// MIX2 tab[55]  high nible=source
//0..5=ch1..ch6 (6..7=VRA..VRB >source only)
{
   	data[MIX1SDOFSET+MIX2OFSET] = incHbyte(data[MIX1SDOFSET+MIX2OFSET]);
	//  c=data[MIX1SDOFSET+MIX2OFSET]/16;
 	lcd_puts_p(MIXCH[data[MIX1SDOFSET+MIX2OFSET]/16]);
}

void func702(void)
// MIX2 tab[55]  low nible=dest

{
   	data[MIX1SDOFSET+MIX2OFSET] = incLbyte(data[MIX1SDOFSET+MIX2OFSET]);
 	lcd_puts_p(MIXCH[data[MIX1SDOFSET+MIX2OFSET] & 0x0F]);
}

void func703(void)
// MIX2 UR
{
    data[MIX1UROFSET+MIX2OFSET] = counter(-100, +100, data[MIX1UROFSET+MIX2OFSET], State, direction); 
    lcd_gotoxy(9,1); 

    lcd_putint(data[MIX1UROFSET+MIX2OFSET],4);
}

void func704(void)
// MIX2 DR
{
    data[MIX1DROFSET+MIX2OFSET] = counter(-100, +100, data[MIX1DROFSET+MIX2OFSET], State, direction); 
    lcd_gotoxy(9,1); 
    lcd_putint(data[MIX1DROFSET+MIX2OFSET],4);
}

void func705(void)
// MIX2 Switch 
{
	data[MIX1SWOFSET+MIX2OFSET] = switchval(data[MIX1SWOFSET+MIX2OFSET]);
	lcd_puts_p(MIXSW[data[MIX1SWOFSET+MIX2OFSET]]);
}


void func801(void)
// MIX3 source
{

   	data[MIX1SDOFSET+MIX3OFSET] = incHbyte(data[MIX1SDOFSET+MIX3OFSET]);   
  	lcd_puts_p(MIXCH[data[MIX1SDOFSET+MIX3OFSET]/16]);
}

void func802(void)
// MIX3 dest
{

	data[MIX1SDOFSET+MIX3OFSET] = incLbyte(data[MIX1SDOFSET+MIX3OFSET]);
    //c=data[MIX1SDOFSET+MIX3OFSET] & 0x0F;
 	lcd_puts_p(MIXCH[data[MIX1SDOFSET+MIX3OFSET] & 0x0F]);
}

void func803(void)
// MIX3 UR +-
{
    data[MIX1UROFSET+MIX3OFSET] = counter(-100, +100, data[MIX1UROFSET+MIX3OFSET], State, direction); 
    lcd_gotoxy(9,1); 

   lcd_putint(data[MIX1UROFSET+MIX3OFSET],4);
}

void func804(void)
// MIX3 DR +-
{
    data[MIX1DROFSET+MIX3OFSET] = counter(-100, +100, data[MIX1DROFSET+MIX3OFSET], State, direction); 
    lcd_gotoxy(9,1); 
    lcd_putint(data[MIX1DROFSET+MIX3OFSET],4);
}

void func805(void)
// MIX3 Switch 
{
	data[MIX1SWOFSET+MIX3OFSET] = switchval(data[MIX1SWOFSET+MIX3OFSET]);
	lcd_puts_p(MIXSW[data[MIX1SWOFSET+MIX3OFSET]]);
}

// SWA var A
void func901(void)
{

        data[SWAVALOFSET]= switchval(data[SWAVALOFSET]);
		lcd_puts_p(SWX[data[SWAVALOFSET]]);

}
// SWB var B
void func1001(void)
{

    data[SWBVALOFSET] = switchval(data[SWBVALOFSET]); 
	lcd_puts_p(SWX[data[SWBVALOFSET]]);

}

void func1101(void)
// SWA pit adj
{
	if (State > 0)    
	{
		State = 0;
		data[SWAPITADJ]++ ;
		if (data[SWAPITADJ] == 2) data[SWAPITADJ]=0; 

	}
	lcd_puts_p(MSV[data[SWAPITADJ]]);

}

void func1201(void)
// SWB pit adj
{
	if (State > 0)    
	{
		State = 0;
		data[SWBPITADJ]++; 
		if (data[SWBPITADJ] ==2) data[SWBPITADJ] = 0; 

	}

	lcd_puts_p(MSV[data[SWBPITADJ]]);
}

int8_t counter(int8_t min, int8_t max, int8_t value, int8_t step, int8_t smer)
{
if (smer == plus) // plus
{
   for (i= 0; i<step; i++)
   {
      if (value < max) value++;
   }  // end for
} // end direction plus
if (smer == minus)  // direction minus
{  
   for ( i=step; i>0 ; i-- )
   {
     if (value > min) value--;
   } //end for

} // end direction minus
	direction = null;
	lcd_putsxy(12,0,"R");
	lcd_putint(State,2);
	// lcd_gotoxy(7,2);
return value;

}


int8_t switchval(int8_t d)
{
   if (State >0)
   {
	  d++;
	  if (d == 4) d=0;
   }
   State = 0;
   return d;
}

int8_t incHbyte(int8_t d)
{
   if (State > 0)
   {
    State = 0;
   	// store H byte
   	c = d/16;
   	// store L byte
   	c1 = d & 0x0F;
   	c++;
 	if (c == 8) c=0;
    d = (c * 16)+ c1;
	}
    return d;
}

int8_t incLbyte(int8_t d)
{
    if (State > 0)
	{
   	State = 0;
   	// store H byte
   	c = d/16;
   	// store L byte
   	c1 = d & 0x0F;
	c1++;
	if (c1 == 8) c1=0;
    d = (c * 16)+ c1;
    }
	return d;
}

int initRadio(void)
{


// start read 17 byte

	// cekani na hlavicku
	for (i = 0; i< 30 ; i++)
	{
		c = USART_readByte();
		data[i] = c;
	}
	// ulozime do EEPROM
/*
    for (i=0; i<30; i++)
	{
		  c = data[i];
	      eeprom_write_byte((uint8_t*)i,c);  
	}  
*/
    i = 0;
	while (1)
	{ 	
      if (checkcrc17(i) == 0) 
	  {
		  LEDPORT |= (1<< LEDOK); 
		  break;	  	
	  }
      i++;
	}

	if (checkcrc17(i) == 0) 
	{
	  	LEDPORT |= (1<< LEDOK); 
	}
    else 
	{
		LEDPORT |= (1<<LEDERR); 
	}
//	lcd_putintxy(0,0,data[16],4);
//	lcd_putintxy(0,1,data[17],4);
//	lcd_putintxy(12,0,i,3);

//    for (i=0; i<18; i++)
//	{
//		  c = data[i];
//	      eeprom_write_byte((uint8_t*)i,c);  
//	}  
// 2. cast natazeni  data 69 byte


return 0;
}

int checkcrc69(void)
{
sum = data[67] * 256;
sum = sum + data[68]; 
for (i=2; i<67; i++)
{
   sum = sum - data[i];
}
return sum;
}

int8_t checkcrc17(int8_t beginIndex)
{
int8_t j;

sum = (uint8_t)data[beginIndex + 16] * 256;
sum = sum + (uint8_t)data[beginIndex + 17]; 

for (j=beginIndex + 2; j<(beginIndex+16); j++)
{
   sum = sum - (uint8_t)data[j];
}
return sum;
}

void USART_writeByte( uint8_t c)
{
   while(!(UCSRA & (1<<UDRE)));
   UDR = c;
//   return 0;
}

void USART_query(void)
{
    USART_writeByte(HEADER);
	USART_writeByte(DOTAZ);
	USART_writeByte(0);

}

uint8_t USART_readByte(void)
{

  // Wait until a byte has been received

  while(!(UCSRA & (1<<RXC))) ;
 
     return UDR;
  // Return received data
}
void SaveToEeprom(void)
{
  // ulozit data69 do eeprom pro pripad resetu a zapamatovani
  //  hodnot
       for (i=0; i<69; i++)
	{
		  c = data[i];
		  // oprava puvodne bez &
	      eeprom_write_byte((uint8_t*)&i,c);  
	} 
}

void LoadFromEeprom(void)
{
	for (i=0; i <69; i++)
	{
	    // oprava puvodne bez &
		data[i] = eeprom_read_byte((uint8_t*)&i);
	}
}

void init_Uart(void)
{
  UBRRH = (((F_CPU/BAUD_RATE)/16)-1)>>8;		// set baud rate
  UBRRL = (((F_CPU/BAUD_RATE)/16)-1);

  /* Enable Tx and Rx */
  UCSRB = (1<<RXEN)|(1<<TXEN);

  /* Set Frame: Data 8 Bit, No Parity and 1 Stop Bit */
  UCSRC = (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0);
  // init LED ports
}

int main(void)
{
	State=0;
    SaveToEeprom();
	// init KBD & LED ports
	KBD_init();
	lcd_init();
	lcd_clear();
	init_Uart();
    // wait until connect RADIO
    lcd_puts_pxy(0,0,INTRO1);
    lcd_puts_pxy(0,1,INTRO2);
     
	initRadio();
    lcd_clear();
	lcd_puts_pxy(0,0,MN000);
	lcd_home();
	// init port
    LoadFromEeprom();
	//Welcome demo message
	 _delay_ms(1500);
    menu_init();
    print_menu();
	Timer0_init();
	while(1)
	{
	 Flag = 1;

	}

}

